/**************************************************************************
 *	Copyright (C) 2004-2007  iCapsid.net
 *
 * 	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version 2
 *	of the License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, 
 *	Boston, MA  02111-1307, USA.
 *
 ****************************************************************************/

#include "fe_include.hpp"
#include "fe_config.hpp"
#include "fe_match.hpp"
#include "fe_os.hpp"
#include "nmap_parser.hpp"

using namespace std;

void NmapParser::startElement(const   XMLCh* const    uri, 
							const   XMLCh* const    localname,
							const   XMLCh* const    qname,
							const   Attributes&		attributes)
{
    char* message = XMLString::transcode(localname);

	if(strcmp(message, "nmaprun") == 0)
	{
		m_Valid = true;
	}
	
	if(!m_Valid){
		throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file"));
		return;
	}
	
	if(strcmp(message, "host") == 0)
	{
		m_CurHost = m_Node->getNewHost();
	}
	else if(strcmp(message, "port") == 0)
	{
		if(!m_CurHost) cerr<<"ERROR"<<endl;

		char* prt = XMLString::transcode(attributes.getValue(XMLString::transcode("protocol")));
	    char* port = XMLString::transcode(attributes.getValue(XMLString::transcode("portid")));

		PortProto proto;

		if(strcmp(prt, "tcp") == 0)
			proto = pp_TCP;
		else // if(strcmp(prt, "udp") == 0)
			proto = pp_UDP;

		m_CurPort = new FePort(proto);
		m_CurPort->setPort(port);
		m_CurHost->addPort(m_CurPort);

		XMLString::release(&port);
		XMLString::release(&prt);
	}
	else if(strcmp(message, "address") == 0)
	{
		if(!m_CurHost){
			throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file: address without host"));
		}

		char* addrt = XMLString::transcode(attributes.getValue(XMLString::transcode("addrtype")));
	    char* addr = XMLString::transcode(attributes.getValue(XMLString::transcode("addr")));
		AddrType addrType;

		if(strcmp(addrt, "mac") == 0)
			addrType = at_MAC;
		else if(strcmp(addrt, "ipv4") == 0)
			addrType = at_IPV4;
		else // if(strcmp(addrt, "ipv6") == 0)
			addrType = at_IPV6;

		m_CurHost->addAddress(new FeAddr(addrType, addr));
		XMLString::release(&addr);
		XMLString::release(&addrt);
	}
	else if(strcmp(message, "service") == 0)
	{
		if(!m_CurPort){
			throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file: service without port"));
		}

		char* n = XMLString::transcode(attributes.getValue(XMLString::transcode("name")));
		m_CurPort->setService(n);
		XMLString::release(&n);
	}
	else if(strcmp(message, "hostname") == 0)
	{
		if(!m_CurHost){
			throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file: hostname without host"));
		}

		char* name = XMLString::transcode(attributes.getValue(XMLString::transcode("name")));
		char* type = XMLString::transcode(attributes.getValue(XMLString::transcode("type")));
	    
		m_CurHost->addHostName(name);

		XMLString::release(&type);
		XMLString::release(&name);

	}
	else if(strcmp(message, "osclass") == 0)
	{
		if(!m_CurHost){
			throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file: osclass without host"));
		}
		char* type = XMLString::transcode(attributes.getValue(XMLString::transcode("type")));
	    char* vendor = XMLString::transcode(attributes.getValue(XMLString::transcode("vendor")));
	    char* family = XMLString::transcode(attributes.getValue(XMLString::transcode("osfamily")));
	    char* acc = XMLString::transcode(attributes.getValue(XMLString::transcode("accuracy")));
	    char* gen = XMLString::transcode(attributes.getValue(XMLString::transcode("osgen")));

		FeOSClass *os = new FeOSClass(type, vendor, family, atoi(acc));

		if(gen) os->setGen(gen);

		m_CurHost->addOSClass(os);

		XMLString::release(&gen);
		XMLString::release(&acc);
		XMLString::release(&family);
		XMLString::release(&vendor);
		XMLString::release(&type);
	}
	else if(strcmp(message, "osmatch") == 0)
	{
		if(!m_CurHost){
			throw(new SAXNotRecognizedException((XMLCh*)L"Invalid nmap XML scan file: osmatch without host"));
		}

		char* name = XMLString::transcode(attributes.getValue(XMLString::transcode("name")));
	    char* acc = XMLString::transcode(attributes.getValue(XMLString::transcode("accuracy")));
	    char* line = XMLString::transcode(attributes.getValue(XMLString::transcode("line")));

		FeOSMatch *os = new FeOSMatch(name, atoi(acc));

		if(line) os->setLine(atoi(line));

		m_CurHost->addOSMatch(os);

		XMLString::release(&line);
		XMLString::release(&acc);
		XMLString::release(&name);
	}

	XMLString::release(&message);
}

void NmapParser::endElement( const XMLCh* const uri,
					 const XMLCh* const localname,
					 const XMLCh* const qname)
{
    char* message = XMLString::transcode(localname);

	if(strcmp(message, "host") == 0)
	{
		FeMatch *tmpMatch = 0; 
		FeMatchParser *matcher = FeMatchParser::instance();
		vector<FeMatch*>* matches = matcher->findMatch(m_CurHost);
		
		if(matches)
		{
			bool tex = false;
			bool geo = false;
			
			for(unsigned int i = 0; i < matches->size() ; ++i)
			{
				tmpMatch = matches->at(i);
				
				if(tmpMatch->getTexture().size() > 0)
				{
					m_CurHost->setTextureID(tmpMatch->getTexture());
					tex = true;	
				}
				
				if(tmpMatch->getGeometry().size() > 0)
				{
					m_CurHost->setGeometryID(tmpMatch->getGeometry());
					geo = true;	
				}			
				
				if(tex && geo) break;
			}
			
			delete matches;
		}
		else
		{
			m_CurHost->setTextureID("unknown");
			m_CurHost->setGeometryID("unknown");
		}

		m_Node->updateRadius();
		m_CurHost = 0;
	}

	XMLString::release(&message);
}

void NmapParser::endDocument()
{

}

void NmapParser::fatalError(const SAXParseException& exception)
{
    char* message = XMLString::transcode(exception.getMessage());
    cerr << "Fatal Error: " << message
         << " at line: " << exception.getLineNumber()
         << endl;
         
	throw(new SAXNotRecognizedException(exception.getMessage()));
    XMLString::release(&message);
}

NmapParser::NmapParser(FeHostsNode *node)
 : m_Valid(false)
{
	m_CurHost 	= 0;
	m_CurPort 	= 0;
	m_Node		= node;
}

NmapParser::~NmapParser()
{
	
}
